from bleach.sanitizer import Cleaner
from cached_property import cached_property
from datetime import timedelta
from dateutil.parser import parse
from onegov.gis import Coordinates
from sedate import replace_timezone


class GuidleBase(object):
    """ Base class for parsing guidle exports containing general helpers. """

    def __init__(self, root):
        self.root = root
        self.nsmap = {'guidle': 'http://www.guidle.com'}
        self.cleaner = Cleaner(
            tags=[],
            attributes={},
            styles=[],
            strip=True,
            filters=[]
        )

    def find(self, path, root=None):
        """ Gets the elements with the given path. """

        root = root if root is not None else self.root
        return root.xpath(path, namespaces=self.nsmap) or []

    def get(self, path, root=None, joiner=' ', parser=None):
        """ Returns the text of the elements with the given path.

        Allows to specifiy a joining character and optionally a parser. If no
        parser is given, the text is HTML-cleaned.

        """

        result = joiner.join([
            (element.text or '').strip() for element in self.find(path, root)
        ]).strip()
        if parser and result:
            result = parser(result)
        else:
            result = self.cleaner.clean(result or '')
        return result

    def join(self, texts, joiner=', '):
        """ Joins a set of text, skips duplicate and empty texts. """

        return joiner.join((
            text for index, text in enumerate(texts)
            if text and text not in texts[index + 1:]
        ))


class GuidleExportData(GuidleBase):
    """ Represents a whole guidle export. """

    def offers(self):
        for offer in self.find('.//guidle:offer'):
            yield GuidleOffer(offer)


class GuidleOffer(GuidleBase):
    """ Represents a single offer containing some description and dates. """

    @cached_property
    def uid(self):
        return self.root.get('id')

    @cached_property
    def last_update(self):
        return self.get('guidle:lastUpdateDate')

    @cached_property
    def title(self):
        title = self.get('guidle:offerDetail/guidle:title')

        # titles are rendered as unsafe html downstream, so we can
        # losen the rules one tiny bit here.

        # we'll still have '&gt,' and '&lt;', but those are probably
        # used very rarely in a title, so we can ignore that
        title = title.replace('&amp;', '&')

        return title

    @cached_property
    def description(self):
        short = self.get('guidle:offerDetail/guidle:shortDescription')
        long = self.get('guidle:offerDetail/guidle:longDescription')
        if long.startswith(short):
            short = ''

        return self.join((
            short,
            long,
            self.get('guidle:offerDetail/guidle:openingHours'),
            self.get('guidle:offerDetail/guidle:externalLink'),
            self.get('guidle:offerDetail/guidle:homepage'),
            self.get('guidle:offerDetail/guidle:priceInformation'),
            self.get('guidle:offerDetail/guidle:ticketingUrl')
        ), '\n\n')

    @cached_property
    def organizer(self):
        return self.join((
            self.get('guidle:contact/guidle:company'),
            self.get('guidle:contact/guidle:name'),
            self.get('guidle:contact/guidle:email'),
            self.get('guidle:contact/guidle:telephone_1'),
        ))

    @cached_property
    def location(self):
        return self.join((
            self.get('guidle:address/guidle:company'),
            self.get('guidle:address/guidle:street'),
            ' '.join((
                self.get('guidle:address/guidle:zip'),
                self.get('guidle:address/guidle:city'),
            )).strip(),
        ))

    def image_url(self, size):
        """ Returns the image url for the offer with the given size or None.

        """

        xpath = (
            f"guidle:offerDetail/"
            f"guidle:images/"
            f"guidle:image/"
            f"guidle:size[@label='{size}']"
        )

        images = self.find(f"({xpath})[1]")

        if not len(images):
            return None

        return images[0].attrib['url']

    def tags(self, tagmap=None):
        """ Returns a set of known and a set of unkonwn tags. """

        tags = self.find(
            'guidle:classifications/'
            'guidle:classification[@type="PRIMARY"]/'
            'guidle:tag'
        )
        tags = [tag.get('subcategoryName') or tag.get('name') for tag in tags]
        tags = set([tag for tag in tags if tag])
        if tagmap:
            return (
                {tagmap[tag] for tag in tags if tag in tagmap},
                tags - tagmap.keys()
            )
        return tags, set()

    @cached_property
    def coordinates(self):
        lat = self.get('guidle:address/guidle:latitude', parser=float)
        lon = self.get('guidle:address/guidle:longitude', parser=float)
        if lat and lon:
            return Coordinates(lat, lon)

    def schedules(self):
        for schedule in self.find('guidle:schedules/guidle:date'):
            yield GuidleScheduleDate(schedule)


class GuidleScheduleDate(GuidleBase):
    """ Represents a single schedule date of an offer. """

    def __init__(self, root):
        super(GuidleScheduleDate, self).__init__(root)

        #  Parse start date, end date and recurrence
        start = self.get('guidle:startDate', parser=parse)
        end = self.get('guidle:endDate', parser=parse)
        recurrence = ''
        weekdays = self.get('guidle:weekdays/guidle:day', joiner=',')
        if weekdays:
            recurrence = f'RRULE:FREQ=WEEKLY;BYDAY={weekdays.upper()}'
        if end and not recurrence:
            if start != end:
                until = end + timedelta(days=1)
                recurrence = (
                    f'RRULE:FREQ=WEEKLY;'
                    f'BYDAY=MO,TU,WE,TH,FR,SA,SU;'
                    f'UNTIL={until:%Y%m%dT%H%MZ}'
                )
                end = start
        else:
            if recurrence:
                until = end + timedelta(days=1)
                recurrence += f';UNTIL={until:%Y%m%dT%H%MZ}'
            end = start

        # Parse start and end times
        start_time = self.get(
            'guidle:startTime', parser=lambda x: parse(x, default=start)
        )
        end_time = self.get(
            'guidle:endTime', parser=lambda x: parse(x, default=end)
        )
        if start_time == end_time and recurrence:
            start_time = None
            end_time = None
        if start_time:
            start = start_time
        if end_time:
            end = end_time
        if end <= start:
            end = end + timedelta(days=1) - timedelta(microseconds=1)

        # Add timezones
        if start:
            start = replace_timezone(start, self.timezone)
        if end:
            end = replace_timezone(end, self.timezone)

        self.start = start
        self.end = end
        self.recurrence = recurrence

    @cached_property
    def timezone(self):
        return 'Europe/Zurich'



import logging
log = logging.getLogger('onegov.event')  # noqa
log.addHandler(logging.NullHandler())  # noqa

from onegov.event.models import Event, Occurrence
from onegov.event.collections import EventCollection, OccurrenceCollection


__all__ = [
    'Event',
    'EventCollection',
    'log',
    'Occurrence',
    'OccurrenceCollection',
]



from datetime import datetime
from icalendar import Calendar as vCalendar
from icalendar import Event as vEvent
from onegov.core.orm import Base
from onegov.core.orm.mixins import TimestampMixin
from onegov.core.orm.types import UUID
from onegov.event.models.mixins import OccurrenceMixin
from pytz import UTC
from sedate import to_timezone
from sqlalchemy import Column
from sqlalchemy import ForeignKey
from uuid import uuid4


class Occurrence(Base, OccurrenceMixin, TimestampMixin):
    """ Defines an occurrence of an event. """

    __tablename__ = 'event_occurrences'

    #: Internal number of the occurence
    id = Column(UUID, primary_key=True, default=uuid4)

    #: Event this occurrence belongs to
    event_id = Column(UUID, ForeignKey('events.id'), nullable=False)

    def as_ical(self, url=None):
        """ Returns the occurrence as iCalendar string. """

        modified = self.modified or self.created or datetime.utcnow()
        event = self.event

        vevent = vEvent()
        vevent.add('uid', f'{self.name}@onegov.event')
        vevent.add('summary', self.title)
        vevent.add('dtstart', to_timezone(self.start, UTC))
        vevent.add('dtend', to_timezone(self.end, UTC))
        vevent.add('last-modified', modified)
        vevent.add('dtstamp', modified)
        vevent.add('location', self.location)
        vevent.add('description', event.description)
        vevent.add('categories', event.tags)
        if event.coordinates:
            vevent.add('geo', (event.coordinates.lat, event.coordinates.lon))
        if url:
            vevent.add('url', url)

        vcalendar = vCalendar()
        vcalendar.add('prodid', '-//OneGov//onegov.event//')
        vcalendar.add('version', '2.0')
        vcalendar.add_component(vevent)
        return vcalendar.to_ical()



from datetime import datetime
from dateutil.parser import parse
from lxml import etree
from onegov.core.utils import module_path
from onegov.event.utils import as_rdates
from onegov.event.utils import GuidleExportData
from pytest import mark
from pytest import raises
from sedate import replace_timezone


def tzdatetime(year, month, day, hour, minute, seconds=0, microseconds=0):
    return replace_timezone(
        datetime(year, month, day, hour, minute, seconds, microseconds),
        timezone='Europe/Zurich'
    )


@mark.parametrize("xml", [
    module_path('onegov.event', 'tests/fixtures/guidle.xml'),
])
def test_import_guidle(session, xml):
    offers = list(GuidleExportData(etree.parse(xml)).offers())
    assert len(offers) == 1

    assert offers[0].uid == '551262854'
    assert offers[0].title == "Theatervorstellung"
    assert offers[0].last_update == "2017-10-21T22:44:12.834+02:00"
    assert offers[0].description == (
        "Lorem ipsum\n\n"
        "Lorem ipsum dolor sit amet, consetetur sadipscing elitr, sed diam "
        "nonumy eirmod tempor invidunt ut labore et dolore magna "
        "aliquyam.\n\n"
        "anschliessend Apéro\n\n"
        "1-Tages- oder 2-Tagespass erhältlich\n\n"
        "www.theatermenzingen.ch"
    )
    assert offers[0].organizer == (
        "Theatervereinigung Menzingen, Peter Muster, "
        "peter.muster@theatermenzingen.ch, +41 41 755 35 46"
    )
    assert offers[0].location == (
        "Zentrum Schützenmatt, Luegetenstrasse 3, 6313 Menzingen"
    )
    assert int(offers[0].coordinates.lat) == 47
    assert int(offers[0].coordinates.lon) == 8
    assert offers[0].tags() == (
        {'Konzert Pop / Rock / Jazz', 'Kulinarik'},
        set()
    )
    assert offers[0].tags({'Konzert Pop / Rock / Jazz': 'Konzert'}) == (
        {'Konzert'}, {'Kulinarik'}
    )

    schedules = list(offers[0].schedules())
    assert len(schedules) == 4

    assert schedules[0].start == tzdatetime(2018, 10, 26, 20, 0)
    assert schedules[0].end == tzdatetime(2018, 10, 26, 22, 30)
    assert schedules[0].recurrence == (
        'RRULE:FREQ=WEEKLY;BYDAY=TU,FR,SA;UNTIL=20181104T0000Z'
    )

    assert schedules[1].start == tzdatetime(2018, 10, 28, 17, 0)
    assert schedules[1].end == tzdatetime(2018, 10, 28, 23, 59, 59, 999999)
    assert schedules[1].recurrence == ''

    assert schedules[2].start == tzdatetime(2018, 6, 15, 8, 0)
    assert schedules[2].end == tzdatetime(2018, 6, 15, 23, 59, 59, 999999)
    assert schedules[2].recurrence == (
        'RRULE:FREQ=WEEKLY;BYDAY=MO,TU,WE,TH,FR,SA,SU;UNTIL=20181016T0000Z'
    )

    assert schedules[3].start == tzdatetime(2018, 8, 18, 0, 0)
    assert schedules[3].end == tzdatetime(2018, 8, 18, 23, 59, 59, 999999)
    assert schedules[3].recurrence == (
        'RRULE:FREQ=WEEKLY;BYDAY=MO,TU,WE,TH,FR,SA,SU;UNTIL=20181101T0000Z'
    )


def test_as_rdates():
    with raises(AssertionError):
        as_rdates('RRULE:FREQ=DAILY;COUNT=1\nRDATE:20190223T000000\n')

    rrule = (
        'RRULE:FREQ=DAILY;COUNT=1\n'
        'RDATE:20190223T000000\n'
    )
    assert as_rdates(rrule, parse('2018-11-03T09:00:00+00:00')) == (
        'RDATE:2018-11-03T000000Z\n'
        'RDATE:2019-02-23T000000Z'
    )

    rrule = (
        'RRULE:FREQ=DAILY;COUNT=7\n'
        'EXDATE:20181111T000000,20181112T000000,20181113T000000,'
        '20181114T000000,20181115T000000,20181116T000000\n'
        'RDATE:20181201T000000,20190112T000000\n'
    )
    assert as_rdates(rrule, parse('2018-09-05T10:30:00+00:00')) == (
        'RDATE:2018-09-05T000000Z\n'
        'RDATE:2018-09-06T000000Z\n'
        'RDATE:2018-09-07T000000Z\n'
        'RDATE:2018-09-08T000000Z\n'
        'RDATE:2018-09-09T000000Z\n'
        'RDATE:2018-09-10T000000Z\n'
        'RDATE:2018-09-11T000000Z\n'
        'RDATE:2018-12-01T000000Z\n'
        'RDATE:2019-01-12T000000Z'
    )

    rrule = (
        'RRULE:FREQ=MONTHLY;BYDAY=-1FR;COUNT=6\n'
        'EXDATE:20180727T000000,20181224T000000,20181228T000000,'
        '20190125T000000\n'
        'RDATE:20181207T000000,20181222T000000\n'
    )
    assert as_rdates(rrule, parse('2018-08-31T07:30:00+00:00')) == (
        'RDATE:2018-08-31T000000Z\n'
        'RDATE:2018-09-28T000000Z\n'
        'RDATE:2018-10-26T000000Z\n'
        'RDATE:2018-11-30T000000Z\n'
        'RDATE:2018-12-07T000000Z\n'
        'RDATE:2018-12-22T000000Z'
    )

    rrule = (
        'RRULE:FREQ=MONTHLY;BYDAY=-1SA;COUNT=4\n'
        'EXDATE:20181027T000000,20181124T000000\n'
        'RDATE:20181110T000000,20181215T000000\n'
    )
    assert as_rdates(rrule, parse('2018-08-25T08:30:00+00:00')) == (
        'RDATE:2018-08-25T000000Z\n'
        'RDATE:2018-09-29T000000Z\n'
        'RDATE:2018-11-10T000000Z\n'
        'RDATE:2018-12-15T000000Z'
    )

    rrule = (
        'RRULE:FREQ=MONTHLY;BYDAY=+1FR;UNTIL=20181207T000000\n'
    )
    assert as_rdates(rrule, parse('2018-01-05T15:00:00+00:00')) == (
        'RDATE:2018-01-05T000000Z\n'
        'RDATE:2018-02-02T000000Z\n'
        'RDATE:2018-03-02T000000Z\n'
        'RDATE:2018-04-06T000000Z\n'
        'RDATE:2018-05-04T000000Z\n'
        'RDATE:2018-06-01T000000Z\n'
        'RDATE:2018-07-06T000000Z\n'
        'RDATE:2018-08-03T000000Z\n'
        'RDATE:2018-09-07T000000Z\n'
        'RDATE:2018-10-05T000000Z\n'
        'RDATE:2018-11-02T000000Z\n'
        'RDATE:2018-12-07T000000Z'
    )

    rrule = (
        'RRULE:FREQ=MONTHLY;BYDAY=+4SA;COUNT=5\n'
        'EXDATE:20181124T000000,20181222T000000\n'
        'RDATE:20181117T000000,20181215T000000\n'
    )
    as_rdates(rrule, parse('2018-08-25T13:00:00+00:00')) == (
        'RDATE:2018-08-25T000000Z\n'
        'RDATE:2018-09-22T000000Z\n'
        'RDATE:2018-10-27T000000Z\n'
        'RDATE:2018-11-17T000000Z\n'
        'RDATE:2018-12-15T000000Z'
    )



from dateutil.rrule import rrulestr
from sedate import standardize_date
from datetime import datetime


def as_rdates(recurrence, dtstart=None):
    assert dtstart or 'DTSTART' in recurrence
    if isinstance(dtstart, datetime):
        dates = list(rrulestr(recurrence, dtstart=dtstart.date()))
    else:
        dates = list(rrulestr(recurrence, dtstart=dtstart))
    dates = [standardize_date(date, timezone='UTC') for date in dates]
    return '\n'.join(f"RDATE:{d.strftime('%Y-%m-%dT%H%M%SZ')}" for d in dates)



from cached_property import cached_property
from datetime import date
from datetime import timedelta
from dateutil.relativedelta import relativedelta
from icalendar import Calendar as vCalendar
from onegov.core.collection import Pagination
from onegov.core.utils import get_unique_hstore_keys
from onegov.event.models import Event
from onegov.event.models import Occurrence
from sedate import as_datetime
from sedate import replace_timezone
from sedate import standardize_date
from sqlalchemy import and_
from sqlalchemy import distinct
from sqlalchemy import or_
from sqlalchemy.dialects.postgresql import array


class OccurrenceCollection(Pagination):
    """ Manages a list of occurrences.

    Occurrences are read only (no ``add`` method here), they are generated
    automatically when adding a new event.

    Occurrences can be filtered by relative date ranges (``today``,
    ``tomorrow``, ``weekend``, ``week``, ``month``) or by a given start date
    and/or end date. The range filter is dominant and overwrites the start and
    end dates if provided.

    By default, only current occurrences are used.

    Occurrences can be additionally filtered by tags and locations.

    """

    date_ranges = ('today', 'tomorrow', 'weekend', 'week', 'month')

    def __init__(
        self,
        session,
        page=0,
        range=None,
        start=None,
        end=None,
        outdated=False,
        tags=None,
        locations=None
    ):
        self.session = session
        self.page = page
        self.range = range if range in self.date_ranges else None
        self.start, self.end = self.range_to_dates(range, start, end)
        self.outdated = outdated
        self.tags = tags if tags else []
        self.locations = locations if locations else []

    def __eq__(self, other):
        return self.page == other.page

    def subset(self):
        return self.query()

    @property
    def page_index(self):
        return self.page

    def page_by_index(self, index):
        return self.__class__(
            self.session,
            page=index,
            range=self.range,
            start=self.start,
            end=self.end,
            outdated=self.outdated,
            tags=self.tags,
            locations=self.locations
        )

    def range_to_dates(self, range, start=None, end=None):
        """ Returns the start and end date for the given range relative to now.
        Defaults to the given start and end date.

        Valid ranges are:
            - today
            - tomorrow
            - weekend (next or current Friday to Sunday)
            - week (current Monday to Sunday)
            - month (current)

        """
        if range not in self.date_ranges:
            return start, end

        today = date.today()
        weekday = today.weekday()

        if range == 'today':
            return today, today
        if range == 'tomorrow':
            tomorrow = today + timedelta(days=1)
            return tomorrow, tomorrow
        if range == 'weekend':
            return (
                today + timedelta(days=4 - weekday),
                today + timedelta(days=6 - weekday)
            )
        if range == 'week':
            return (
                today - timedelta(days=weekday),
                today + timedelta(days=6 - weekday)
            )
        if range == 'month':
            start = today.replace(day=1)
            return start, start + relativedelta(months=1, days=-1)
        pass

    def for_filter(self, **kwargs):
        """ Returns a new instance of the collection with the given filters
        and copies the current filters if not specified.

        If a valid range is provided, start and end dates are ignored. If the
        range is invalid, it is ignored.

        Adds or removes a single tag/location if given.
        """

        range = kwargs.get('range', self.range)
        start = kwargs.get('start', self.start)
        end = kwargs.get('end', self.end)
        if 'range' in kwargs and kwargs.get('range') in self.date_ranges:
            range = kwargs.get('range')
            start = None
            end = None
        elif 'start' in kwargs or 'end' in kwargs:
            range = None

        tags = kwargs.get('tags', list(self.tags))
        if 'tag' in kwargs:
            tag = kwargs.get('tag')
            if tag in tags:
                tags.remove(tag)
            elif tag is not None:
                tags.append(tag)

        locations = kwargs.get('locations', list(self.locations))
        if 'location' in kwargs:
            location = kwargs.get('location')
            if location in locations:
                locations.remove(location)
            elif location is not None:
                locations.append(location)

        return self.__class__(
            self.session,
            page=0,
            range=range,
            start=start,
            end=end,
            outdated=kwargs.get('outdated', self.outdated),
            tags=tags,
            locations=locations
        )

    @cached_property
    def used_timezones(self):
        """ Returns a list of all the timezones used by the occurrences. """

        return [
            tz[0] for tz in self.session.query(distinct(Occurrence.timezone))
        ]

    @cached_property
    def used_tags(self):
        """ Returns a list of all the tags used by the occurrences.

        This could be solve possibly more effienciently with the skey function
        currently not supported by SQLAlchemy (e.g.
        ``select distinct(skeys(tags))``), see
        http://stackoverflow.com/q/12015942/3690178

        """

        return get_unique_hstore_keys(self.session, Occurrence._tags)

    def query(self):
        """ Queries occurrences with the set parameters.

        Finds occurrences which:
        * are between start and end date
        * have any of the tags
        * have any of the locations (exact word)

        Start and end date are assumed to be dates only and therefore without
        a timezone - we search for the given date in the timezone of the
        occurrence.

        """

        query = self.session.query(Occurrence)

        if self.start is not None or self.outdated is False:
            if self.start is None:
                start = date.today()
            else:
                start = self.start
                if self.outdated is False:
                    start = max(self.start, date.today())
            start = as_datetime(start)

            expressions = []
            for timezone in self.used_timezones:
                localized_start = replace_timezone(start, timezone)
                localized_start = standardize_date(localized_start, timezone)
                expressions.append(
                    and_(
                        Occurrence.timezone == timezone,
                        Occurrence.start >= localized_start
                    )
                )

            query = query.filter(or_(*expressions))

        if self.end is not None:
            end = as_datetime(self.end)
            end = end + timedelta(days=1)

            expressions = []
            for timezone in self.used_timezones:
                localized_end = replace_timezone(end, timezone)
                localized_end = standardize_date(localized_end, timezone)
                expressions.append(
                    and_(
                        Occurrence.timezone == timezone,
                        Occurrence.end < localized_end
                    )
                )

            query = query.filter(or_(*expressions))

        if self.tags:
            query = query.filter(Occurrence._tags.has_any(array(self.tags)))

        if self.locations:
            query = query.filter(
                or_(*[
                    Occurrence.location.op('~')(f'\\y{location}\\y')
                    for location in self.locations
                ])
            )

        query = query.order_by(Occurrence.start, Occurrence.title)

        return query

    def by_name(self, name):
        """ Returns an occurrence by its URL-friendly name.

        The URL-friendly name is automatically constructed as follows:

        ``unique name of the event``-``date of the occurrence``

        e.g.

        ``squirrel-park-visit-6-2015-06-20``

        """

        query = self.session.query(Occurrence).filter(Occurrence.name == name)
        return query.first()

    def as_ical(self, request):
        """ Returns the the events of the given occurrences as iCalendar
        string.

        """

        vcalendar = vCalendar()
        vcalendar.add('prodid', '-//OneGov//onegov.event//')
        vcalendar.add('version', '2.0')

        query = self.query().with_entities(Occurrence.event_id)
        event_ids = set([r.event_id for r in query])

        query = self.session.query(Event).filter(Event.id.in_(event_ids))
        for event in query:
            for vevent in event.get_ical_vevents(request.link(event)):
                vcalendar.add_component(vevent)

        return vcalendar.to_ical()



import click
import pycurl

from csv import reader as csvreader
from datetime import date
from datetime import datetime
from datetime import timedelta
from dateutil.parser import parse
from hashlib import sha1
from icalendar import Calendar as vCalendar
from io import BytesIO
from lxml import etree
from onegov.core.cli import abort
from onegov.core.cli import command_group
from onegov.core.cli import pass_group_context
from onegov.event import log
from onegov.event.collections import EventCollection
from onegov.event.collections.events import EventImportItem
from onegov.event.models import Event
from onegov.event.models import EventFile
from onegov.event.models import Occurrence
from onegov.event.utils import as_rdates
from onegov.event.utils import GuidleExportData
from onegov.file import DepotApp
from onegov.file.utils import as_fileintent
from onegov.gis import Coordinates
from operator import add
from pathlib import Path
from pytz import UTC
from requests import get
from sedate import as_datetime
from sedate import replace_timezone
from sedate import standardize_date
from sedate import to_timezone
from sqlalchemy import or_
from sqlalchemy.dialects.postgresql import array
from urllib.parse import urlparse


cli = command_group()


@cli.command('clear')
@pass_group_context
def clear(group_context):
    """ Deletes all events.

        onegov-event --select '/veranstaltungen/zug' clear

    """

    def _clear(request, app):
        if not click.confirm("Do you really want to remove all events?"):
            abort("Deletion process aborted")

        session = app.session()
        for event in session.query(Event):
            session.delete(event)
        for occurrence in session.query(Occurrence):
            session.delete(occurrence)

    return _clear


def download_image(image_url):
    buffer = BytesIO()

    c = pycurl.Curl()
    c.setopt(c.URL, image_url)
    c.setopt(c.WRITEDATA, buffer)
    c.setopt(c.FOLLOWLOCATION, True)

    try:
        c.perform()

        if c.getinfo(c.RESPONSE_CODE) != 200:
            return None

    finally:
        c.close()

    return buffer


def get_event_dates(url, timezone):
    """ Get the start and end datetime of an event in a seantis.dir.events
    events calendar.

    The start and end datetime in the seantis.dir.events JSON export is
    wrong since they refer to the next occurrence and not the first one. This
    is unusable together with the recurrece

    """

    response = get(urlparse(url)._replace(query='type=ical').geturl())
    response.raise_for_status()

    for event in vCalendar.from_ical(response.text).walk('vevent'):
        start = event.get('dtstart').dt
        if type(start) is date:
            start = replace_timezone(as_datetime(start), timezone)
        elif type(start) is datetime:
            if start.tzinfo is None:
                start = standardize_date(start, timezone)
            else:
                start = to_timezone(start, UTC)

        end = event.get('dtend').dt
        if type(end) is date:
            end = replace_timezone(as_datetime(end), timezone)
            end = end + timedelta(days=1, minutes=-1)
        elif type(end) is datetime:
            if end.tzinfo is None:
                end = standardize_date(end, timezone)
            else:
                end = to_timezone(end, UTC)

        return start, end


@cli.command('import-json')
@pass_group_context
@click.argument('url')
@click.option('--tagmap', type=click.File())
@click.option('--clear/-no-clear', default=False)
def import_json(group_context, url, tagmap, clear):
    """ Fetches the events from a seantis.dir.events instance.

    This command is intended for migration and to be removed in the future.

    Example:

        onegov-event --select '/veranstaltungen/zug' import-json \
        'https://veranstaltungen.zug.ch/veranstaltungen/?type=json&compact'

    """
    if tagmap:
        tagmap = {row[0]: row[1] for row in csvreader(tagmap)}

    def _import_json(request, app):
        unknown_tags = set()

        response = get(url)
        response.raise_for_status()
        response = response.json()

        session = app.session()
        events = EventCollection(session)

        if clear:
            click.secho("Removing all events", fg='yellow')
            session = app.session()
            for event in session.query(Event):
                session.delete(event)
            for occurrence in session.query(Occurrence):
                session.delete(occurrence)

        for item in response:
            title = item['title']

            start = parse(item['start'])
            end = parse(item['end'])
            timezone = item['timezone']
            recurrence = item['recurrence']
            if recurrence:
                start, end = get_event_dates(item['url'], timezone)

            organizer = ', '.join((line for line in (
                item['organizer'] or '',
                item['contact_name'] or '',
                item['contact_email'] or '',
                item['contact_phone'] or ''
            ) if line))

            location = ', '.join((line for line in (
                item['locality'] or '',
                ' '.join((
                    item['street'] or '',
                    item['housenumber'] or ''
                )).strip(),
                ' '.join((
                    item['zipcode'] or '',
                    item['town'] or '',
                )).strip(),
            ) if line))

            description = '\n\n'.join((line for line in (
                item['short_description'] or '',
                item['long_description'] or '',
                item['event_url'] or '',
                item['location_url'] or '',
                item['prices'],
                item['registration']
            ) if line))

            tags = item['cat1']
            if tagmap and tags:
                unknown_tags |= set(tags) - tagmap.keys()
                tags = {tagmap[tag] for tag in tags if tag in tagmap}

            coordinates = None
            if item['latitude'] and item['longitude']:
                coordinates = Coordinates(
                    lat=item['latitude'],
                    lon=item['longitude']
                )

            event = Event(
                state='initiated',
                name=events._get_unique_name(title),
                title=title,
                start=start,
                end=end,
                timezone=timezone,
                description=description,
                organizer=organizer,
                location=location,
                coordinates=coordinates,
                tags=tags or [],
                meta={'submitter_email': item['submitter_email']},
            )

            # if the recurrence rule is not supported, turn it into a
            # list of rdates
            try:
                event.validate_recurrence('recurrence', recurrence)
            except RuntimeError:
                event.recurrence = as_rdates(recurrence, start)
                if not event.recurrence:
                    raise RuntimeError(f"Could not convert '{recurrence}'")
            else:
                event.recurrence = recurrence

            if item['images']:
                buffer = download_image(item['images'][0]['url'])

                if buffer:
                    filename = item['images'][0]['name'] or 'event-image'
                    event.image = EventFile(
                        name=filename,
                        reference=as_fileintent(buffer, filename)
                    )

            session.add(event)
            event.submit()
            event.publish()

        if unknown_tags:
            unknown_tags = ', '.join([f'"{tag}"' for tag in unknown_tags])
            click.secho(f"Tags not in tagmap: {unknown_tags}!", fg='yellow')

        click.secho(f"Imported {len(response)} events", fg='green')

    return _import_json


@cli.command('import-ical')
@pass_group_context
@click.argument('ical', type=click.File())
def import_ical(group_context, ical):
    """ Imports events from an iCalendar file.

    Example:

        onegov-event --select '/veranstaltungen/zug' import-ical import.ics

    """

    def _import_ical(request, app):
        collection = EventCollection(app.session())
        added, updated, purged = collection.from_ical(ical.read())
        click.secho(
            f"Events successfully imported "
            f"({added} added, {updated} updated, {purged} deleted)",
            fg='green')

    return _import_ical


@cli.command('import-guidle')
@pass_group_context
@click.argument('url')
@click.option('--tagmap', type=click.File())
def import_guidle(group_context, url, tagmap):
    """ Fetches the events from guidle.

    Example:

        onegov-event --select '/veranstaltungen/zug' import-guidle \
        'http://www.guidle.com/xxxx/'

    """
    if tagmap:
        tagmap = {row[0]: row[1] for row in csvreader(tagmap)}

    def _import_guidle(request, app):
        try:
            response = get(url)
            response.raise_for_status()

            unknown_tags = set()
            prefix = 'guidle-{}'.format(sha1(url.encode()).hexdigest()[:10])
            collection = EventCollection(app.session())
            updated = dict(collection.query().with_entities(
                Event.meta['source'].astext,
                Event.meta['source_updated'].astext,
            ))

            root = etree.fromstring(response.text.encode('utf-8'))

            def items(unknown_tags):
                for offer in GuidleExportData(root).offers():
                    source = f'{prefix}-{offer.uid}.0'
                    if offer.last_update == updated.get(source):
                        # Do not download the images and parse all the fields
                        # if nothing has changed since the last import. Only
                        # provide some dummy events with the source set so
                        # that the purging will work!
                        for index, schedule in enumerate(offer.schedules()):
                            yield EventImportItem(
                                event=Event(
                                    source=f'{prefix}-{offer.uid}.{index}',
                                    source_updated=offer.last_update
                                ),
                                image=None,
                                filename=None
                            )
                        continue

                    tags, unknown = offer.tags(tagmap)
                    unknown_tags |= unknown

                    image = None
                    filename = None
                    if isinstance(app, DepotApp):
                        image_url = offer.image_url(size='original')
                        if image_url:
                            image = download_image(image_url)
                            filename = image_url.rsplit('/', 1)[-1]

                    for index, schedule in enumerate(offer.schedules()):
                        event = Event(
                            state='initiated',
                            title=offer.title,
                            start=schedule.start,
                            end=schedule.end,
                            recurrence=schedule.recurrence,
                            timezone=schedule.timezone,
                            description=offer.description,
                            organizer=offer.organizer,
                            location=offer.location,
                            coordinates=offer.coordinates,
                            tags=tags,
                            source=f'{prefix}-{offer.uid}.{index}',
                            source_updated=offer.last_update
                        )

                        if image:
                            image.seek(0)

                        yield EventImportItem(
                            event=event,
                            image=image,
                            filename=filename
                        )

                    if image:
                        image.close()

            added, updated, purged = collection.from_import(
                items(unknown_tags),
                prefix
            )

            if unknown_tags:
                unknown_tags = ', '.join([f'"{tag}"' for tag in unknown_tags])
                click.secho(
                    f"Tags not in tagmap: {unknown_tags}!", fg='yellow'
                )

            click.secho(
                f"Events successfully imported from '{url}' "
                f"({added} added, {updated} updated, {purged} deleted)",
                fg='green'
            )
        except Exception as e:
            log.error("Error importing events", exc_info=True)
            raise(e)

    return _import_guidle


@cli.command()
@pass_group_context
@click.option('--source', multiple=True)
@click.option('--tag', multiple=True)
@click.option('--location', multiple=True)
def fetch(group_context, source, tag, location):
    """ Fetches events from other instances.

    Only fetches events from the same namespace which have not been imported
    themselves.

    Example

        onegov-event --select '/veranstaltungen/zug' fetch \
            --source menzingen --source steinhausen
            --tag Sport --tag Konzert
            --location Zug

    """

    def vector_add(a, b):
        return list(map(add, a, b))

    if not len(source):
        abort("Provide at least one source")

    def _fetch(request, app):

        def event_file(reference):
            # XXX use a proper depot manager for this
            path = Path(app.depot_storage_path) / reference['path'] / 'file'
            with open(path):
                content = BytesIO(path.open('rb').read())
            return content

        try:
            result = [0, 0, 0]

            for key in source:
                remote_schema = '{}-{}'.format(app.namespace, key)
                local_schema = app.session_manager.current_schema
                assert remote_schema in app.session_manager.list_schemas()

                app.session_manager.set_current_schema(remote_schema)
                remote_session = app.session_manager.session()
                assert remote_session.info['schema'] == remote_schema

                query = remote_session.query(Event)
                query = query.filter(
                    or_(
                        Event.meta['source'].astext.is_(None),
                        Event.meta['source'].astext == ''
                    )
                )
                if tag:
                    query = query.filter(Event._tags.has_any(array(tag)))
                if location:
                    query = query.filter(
                        or_(*[
                            Event.location.op('~')(f'\\y{term}\\y')
                            for term in location
                        ])
                    )

                def remote_events():
                    for event in query:
                        event._es_skip = True
                        yield EventImportItem(
                            event=Event(
                                state='initiated',
                                title=event.title,
                                start=event.start,
                                end=event.end,
                                timezone=event.timezone,
                                recurrence=event.recurrence,
                                content=event.content,
                                location=event.location,
                                tags=event.tags,
                                source=f'fetch-{key}-{event.name}',
                                coordinates=event.coordinates,
                            ),
                            image=(
                                event_file(event.image.reference)
                                if event.image else None
                            ),
                            filename=event.image.name if event.image else None
                        )

                # be sure to switch back to the local schema, lest we
                # accidentally update things on the remote
                app.session_manager.set_current_schema(local_schema)
                local_events = EventCollection(app.session_manager.session())

                result = vector_add(
                    result,
                    local_events.from_import(remote_events(), f'fetch-{key}')
                )

            click.secho(
                f"Events successfully fetched "
                f"({result[0]} added, {result[1]} updated, "
                f"{result[2]} deleted)",
                fg='green'
            )

        except Exception as e:
            log.error("Error fetching events", exc_info=True)
            raise(e)

    return _fetch
